@@ -211,7 +211,39 @@ int smtp_open(struct query *ctl)
 		ctl->destaddr = (ctl->smtphost && ctl->smtphost[0] != '/') ? ctl->smtphost : "localhost";
     } 
     else 
-	ctl->destaddr = ctl->smtpaddress ? ctl->smtpaddress : ( ctl->smtphost && ctl->smtphost[0] != '/' ? ctl->smtphost : "localhost");
+      {
+	/* 
+	 * Here we try to find a correct domain name part for the RCPT
+	 * TO address.  If smtpaddress is set, no need to guestimate
+	 * it.  Otherwise, using ctl->smtphost as a base is a good
+	 * base, although we may have to strip any port appended to
+	 * communicate with SMTP servers that do not listen on the
+	 * SMTP port.  (benj) */
+	if (ctl->smtpaddress)
+	  ctl->destaddr = ctl->smtpaddress;
+	else if (ctl->smtphost && ctl->smtphost[0] != '/')
+	  {
+	    char * cp;
+	    if (cp = strchr (ctl->smtphost, '/'))
+	    {
+	      /* As an alternate port for smtphost is specified, we
+		 need to strip it from domain name. */
+	      char *smtpname;
+	      xalloca(smtpname, char *, cp - ctl->smtphost + 1);
+	      strncpy(smtpname, ctl->smtphost, cp - ctl->smtphost +1);
+	      cp = strchr(smtpname, '/');
+	      *cp = 0;
+	      ctl->destaddr = smtpname;
+	    }
+	    else
+	      /* No need to strip port, domain name is smtphost. */
+	      ctl->destaddr = ctl->smtphost;
+	  }
+	/* No smtphost is specified or it is a UNIX socket, then use
+	   localhost as a domain part. */
+	else
+	  ctl->destaddr = "localhost";
+      }
 
     if (outlevel >= O_DEBUG && ctl->smtp_socket != -1)
 	report(stdout, GT_("forwarding to %s\n"), ctl->smtphost);
@@ -390,8 +422,11 @@ static int send_bouncemail(struct query *ctl, struct msgblk *msg,
     SockPrintf(sock, "--%s\r\n", boundary); 
     SockPrintf(sock, "Content-Type: text/rfc822-headers\r\n");
     SockPrintf(sock, "\r\n");
-    SockWrite(sock, msg->headers, strlen(msg->headers));
-    SockPrintf(sock, "\r\n");
+    if (msg->headers)
+    {
+	SockWrite(sock, msg->headers, strlen(msg->headers));
+	SockPrintf(sock, "\r\n");
+    }
     SockPrintf(sock, "--%s--\r\n", boundary); 
 
     if (SMTP_eom(sock) != SM_OK || SMTP_quit(sock))
@@ -689,7 +724,7 @@ static int open_bsmtp_sink(struct query *ctl, struct msgblk *msg,
 
     /* see the ap computation under the SMTP branch */
     fprintf(sinkfp, 
-	    "MAIL FROM: %s", (msg->return_path[0]) ? msg->return_path : user);
+	    "MAIL FROM:%s", (msg->return_path[0]) ? msg->return_path : user);
 
     if (ctl->pass8bits || (ctl->mimemsg & MSG_IS_8BIT))
 	fputs(" BODY=8BITMIME", sinkfp);
@@ -1479,7 +1514,7 @@ int open_warning_by_mail(struct query *ctl, struct msgblk *msg)
     }
     else				/* send to postmaster  */
 	status = open_sink(ctl, &reply, &good, &bad);
-    stuff_warning(ctl, "Date: %s", rfc822timestamp());
+    if (status == 0) stuff_warning(ctl, "Date: %s", rfc822timestamp());
     return(status);
 }
 
